mouse = {}

local localMouse = {}

local mouseInputs = {}
setmetatable(mouse, {
    __index = function(_, key)
        if key == "x" or key == "y" or key == "position" or key == "defaultInterval" or key == "dx" or key == "dy" or key == "inRegion" or key == "active" or key == "hide" or key == "show" or key == "hidden" or key == "left" or key == "right" or key == "middle" or key == "scroll" or key == "init" or key == "version" then
            return localMouse[key]
        end
        return rawget(mouse, key)
    end,
    
    __newindex = function(_,key, value)
        if key == "x" or key == "y" or key == "position" or key == "defaultInterval" or key == "dx" or key == "dy" or key == "inRegion" or key == "active" or key == "hide" or key == "show" or key == "hidden" or key == "left" or key == "right" or key == "middle" or key == "scroll" or key == "init" or key == "version" then
            error("You can not change property \""..key.."\" in mouse")
        elseif key == "interval" or key == "connected" or key == "disconnected" or key == "lock" then
            rawset(mouse, key, value)
        else
            error("Property \""..key.."\" does not exist in mouse")
        end
    end,
    __len = function()
        return #mouseInputs
    end,
    __name = "mouse",
    __tostring = function() return "mouse class: version "..localMouse.version  end
})
localMouse.version = 1.2

for k, gcmouse in ipairs(objc.GCMouse.mice_()) do
    table.insert(mouseInputs, gcmouse.mouseInput)
end

local interaction = nil
function localMouse:init()
    --Pos
    localMouse.x = 0
    localMouse.y = 0
    localMouse.inRegion = false
    localMouse.dx = 0
    localMouse.dy = 0
    localMouse.active = (objc.GCMouse.current ~= nil)
    mouse.lock = false
    localMouse.position = vec2(0)
    localMouse.hidden = false
    mouse.interval = 0.25
    localMouse.defaultInterval = 0.25

    local HoverHandler = objc.class("HoverHandler")
    
    function HoverHandler:hovering(obj)
        hoverPos = obj:locationInView_(obj.view)
        localMouse.x = hoverPos.x
        localMouse.y = HEIGHT-hoverPos.y
    end
    
    hoverHandler = HoverHandler()
    hoverGesture = objc.UIHoverGestureRecognizer:alloc():initWithTarget_action_(hoverHandler, objc.selector("hovering"))
    objc.viewer.view:addGestureRecognizer_(hoverGesture)
    -- Hide

    
    
    PointerInteraction = objc.delegate("UIPointerInteractionDelegate")
    function PointerInteraction:pointerInteraction_regionForRequest_defaultRegion_(objInteraction, objRegion, objDefault)    
        local frame = objc.viewer.view.frame
        localMouse.inRegion = true
        return objc.UIPointerRegion:regionWithRect_identifier_(frame, "ViewerRegion")
    end
    
    
    function PointerInteraction:pointerInteraction_styleForRegion_(objInteraction, objRegion)
        local hidden = objc.UIPointerStyle.hiddenPointerStyle
        return hidden
    end
    
    
    delegate = PointerInteraction()
    interaction = objc.UIPointerInteraction:alloc():initWithDelegate_(delegate)
    
    interaction.delegate = delegate
    
    -- Broken
    --[=[mouseInput.mouseMovedHandler = function(objMouse, floatDeltaX, floatDeltaY)
        mouse.dx = floatDeltaX
        mouse.dy = floatDeltaY
    end]=]
    
end

function localMouse:hide()
    if not mouse.hidden then
        objc.viewer.view :addInteraction_(interaction)
        localMouse.hidden = true
    end
end

function localMouse:show()
    if mouse.hidden then
        objc.viewer.view :removeInteraction_(interaction)
        localMouse.hidden = false
    end
end


function touched(touch)
    if touch.type == touch.pointer then
        localMouse.x = touch.x
        localMouse.y = touch.y
    end
end


NotificationHandler = objc.class("NotificationHandler")
notificationCenter = objc.NSNotificationCenter.defaultCenter

function NotificationHandler:mouseDidConnect(objNotification)
    mouseInputs[#objc.GCMouse.mice_()] = objNotification.object.mouseInput
    setMouseHandlers(mouseInputs[#objc.GCMouse.mice_()])
    if(mouse.connected ~= nil) then
        mouse.connected()
    end
    --[=[mouseInput.mouseMovedHandler = function(objMouse, floatDeltaX, floatDeltaY)
        mouse.dx = floatDeltaX
        mouse.dy = floatDeltaY
    end]=]
end


handler = NotificationHandler()

notificationCenter:removeObserver_name_object_("GCMouseDidConnectNotification")

notificationCenter:addObserver_selector_name_object_(
handler,
objc.selector("mouseDidConnect"),
"GCMouseDidConnectNotification"
)

function NotificationHandler:mouseDidDisconnect(objNotification)
    for i = #mouseInputs, 1, -1 do
        if(mouseInputs[i] == objNotification.object.mouseInput) then
            table.remove(mouseInputs, i)
            break
        end
    end
    if(mouse.disconnected ~= nil) then
        mouse.disconnected()
    end
end


handler2 = NotificationHandler()

notificationCenter:addObserver_selector_name_object_(
handler2,
objc.selector("mouseDidDisconnect"),
"GCMouseDidDisconnectNotification"
)

function shutdown()
    notificationCenter:removeObserver_(handler)
    notificationCenter:removeObserver_(handler2)
end

local justChanged = {}


local mouseTypes = {"left", "middle", "right"}
local checkingClicks = {}
local timeOfLastAction = {}
local numClicks = {}
local timeOfChange = {}
local localMouseButtons = {}

for k, mouseType in ipairs(mouseTypes) do
    localMouseButtons[mouseType] = {pressed = false, released = false, click = false, doubleClick = false, tripleClick = false, numClicks = 0, timePressing = 0, timeLastPressed = 0, timeReleasing = 0, timeLastReleased = 0}

    
    localMouse[mouseType] = setmetatable({}, {
        __index = function(_, key)
            return localMouseButtons[mouseType][key]
        end,
        __newindex = function(_,key, value)
            error("Can not set any property of mouse.button")
        end
    })
    checkingClicks[mouseType] = false
    timeOfLastAction[mouseType] = 0
    numClicks[mouseType] = 0
    timeOfChange[mouseType] = 0
    justChanged[mouseType] = false
end

function mouseHandler(name)
    return function (objButton, floatValue, boolPressed)
        if boolPressed then
            localMouseButtons[name].pressed = true
        else
            localMouseButtons[name].released = true
        end
        justChanged[name] = true
        
    end
end



function setMouseHandlers(mouseInput)
    if mouseInput then
        mouseInput.leftButton.pressedChangedHandler = mouseHandler("left")
        mouseInput.rightButton.pressedChangedHandler = mouseHandler("right")
        mouseInput.middleButton.pressedChangedHandler = mouseHandler("middle")
    end
end
for k, mouseInput in ipairs(mouseInputs) do
    setMouseHandlers(mouseInput)
end

localMouse.scroll = vec2(0)

local function checkForClicks(mouseType)
    --Reset Clicks
    localMouseButtons[mouseType].click = false
    localMouseButtons[mouseType].doubleClick = false
    localMouseButtons[mouseType].tripleClick = false
    localMouseButtons[mouseType].numClicks = 0
        
    if mouse[mouseType].pressed then
        if not checkingClicks[mouseType] then
            checkingClicks[mouseType] = true
        end
        timeOfLastAction[mouseType] = time.elapsed
    end
    if checkingClicks[mouseType] and mouse[mouseType].released then
        numClicks[mouseType] = numClicks[mouseType] + 1
    end
    if timeOfLastAction[mouseType] + mouse.interval < time.elapsed and checkingClicks[mouseType]  then
        checkingClicks[mouseType] = false
        if numClicks[mouseType] == 1 then
            localMouseButtons[mouseType].click = true
        elseif numClicks[mouseType] == 2 then
            localMouseButtons[mouseType].doubleClick = true
        elseif numClicks[mouseType] == 3 then
            localMouseButtons[mouseType].tripleClick = true
        end
        localMouseButtons[mouseType].numClicks = numClicks[mouseType]
        numClicks[mouseType] = 0
    end
end

local function checkMouseTime(mouseType)
    if not mouse[mouseType].pressing then
        localMouseButtons[mouseType].timeReleasing = time.elapsed - timeOfChange[mouseType]
    else
        localMouseButtons[mouseType].timePressing = time.elapsed - timeOfChange[mouseType]
    end
        
    if mouse[mouseType].released then
        localMouseButtons[mouseType].timeLastPressed = mouse[mouseType].timePressing
        localMouseButtons[mouseType].timePressing = 0
        timeOfChange[mouseType] = time.elapsed
    end
    if mouse[mouseType].pressed then
        localMouseButtons[mouseType].timeLastReleased = mouse[mouseType].timeReleasing
        localMouseButtons[mouseType].timeReleasing = 0
        timeOfChange[mouseType] = time.elapsed
    end
end

function update()
    if #mouseInputs >= 1 then
        viewer.pointerLocked = mouse.lock
        localMouse.position = vec2(mouse.x, mouse.y)
        localMouse.scroll.x = 0
        localMouse.scroll.y = 0
        for k, mouseType in ipairs(mouseTypes) do
            localMouseButtons[mouseType].pressing = false
            localMouseButtons[mouseType].value = 0.0
        end
        for k, mouseInput in ipairs(mouseInputs) do
            localMouse.scroll.x = (math.abs(mouseInput.scroll.xAxis.value) > math.abs(localMouse.scroll.x)) and mouseInput.scroll.xAxis.value or localMouse.scroll.x
            localMouse.scroll.y = (math.abs(mouseInput.scroll.yAxis.value) > math.abs(localMouse.scroll.y)) and mouseInput.scroll.yAxis.value or localMouse.scroll.y
            for k, mouseType in ipairs(mouseTypes) do
                localMouseButtons[mouseType].pressing = (mouseInput[mouseType.."Button"].value == 1.0) or localMouseButtons[mouseType].pressing
                localMouseButtons[mouseType].value = localMouseButtons[mouseType].pressing and 1.0 or 0.0
            end  
        end
        for k, mouseType in ipairs(mouseTypes) do
            if not justChanged[mouseType] then
                localMouseButtons[mouseType].pressed = false
                localMouseButtons[mouseType].released = false
            else
                justChanged[mouseType] = false
            end
            checkForClicks(mouseType)
            checkMouseTime(mouseType)
        end
        
        localMouse.inRegion =  not ((mouse.x == 0 and mouse.y == HEIGHT) and not mouse.lock)
        localMouse.active = true
    else
        localMouse.active = false
        localMouse.inRegion = false
    end
    
end
localMouse:init()